/* ------------------------------------------------------------- */
/* T34.cc                                                        */
/* Stefan Bauberger (bauberger@physik.uni-wuerzburg.de)          */
/* Ayres Freitas (afreitas@pitt.edu)			         */
/* last revision 19.04.19 (4.12.95)                              */
/* ------------------------------------------------------------- */
/* Scalar integrals T234, T1234, T11234 for two-loop             */
/* self-energies                                                 */
/* as part of the package s2lse for the calculation of scalar    */
/* two-loop self-energy integrals                                */
/* ------------------------------------------------------------- */
/* References:                                                   */
/*   [1] = Bauberger, Berends, Boehm, Buza,                      */
/*         hep-ph/9409388, Nucl.Phys.B 343 (1995) 383            */
/*   [2] = Bauberger, Boehm,                                     */
/*         hep-ph/9501201, Nucl.Phys.B 445 (1995) 25             */
/*   [3] = Berends, Tausk, Nucl. Phys. B 421 (1994) 456          */
/*   [4] = Scharf, Tausk, Nucl. Phys. B 412 (1994) 423           */
/*   [5] = Broadhurst, Z. Phys. C 47 (1990) 115                  */
/* ------------------------------------------------------------- */

#include "T345.h"
#include "li.h"
#include "qagutil.h"
#include "B0.h"
#include "tinline.h"

#define Errstream cerr

/******************************************************************/
/* some constants which define from which value on the asymptotic */
/* limit of some functions will be taken:                         */
/******************************************************************/

#ifdef QUADPRECISION
static Double T11234ASYMPLIMIT1=SDouble("1e-14");
#else
#define T11234ASYMPLIMIT1 1e-4
#endif

// zeta(n):
#define MAX_ZETA 5
static Double zeta[MAX_ZETA+1] =
{ SDouble("0"),SDouble(" 0"),
  SDouble(" 1.64493406684822643647241516665"),
  SDouble("1.202056903159594285399738161511"),
  SDouble(" 1.08232323371113819151600369654"),
  SDouble("1.036927755143369926331365486457") };

#define Zeta zeta

/******************************************************************/
/* Throughout this file the numbers are defined as Double or Cplx */
/* to enable a generalization to arbitrary precision              */
/******************************************************************/

/******************************************************************/
/* convention: vars is frequently used for (var)^2                */
/* for example: m1s denotes m_1^2.                                */
/******************************************************************/

static Double ps,m1s,m2s,m3s,m4s,threshold;
static Cplx T234sub0k(Double x);
static Cplx T1234C3sub0kp(Double x);
static Cplx T11234C3M1Zsub0k(Double s);
static Cplx T11234fin0kp2(Double x);
static Cplx T11234fin0kDp2(Double x);

static Cplx T234sub0k(Double s)
{
  return(-DeltaB0OnTwoPiI(s,m2s,m3s)
           *(B0D(ps,m1s,s)-B0D(0,m1s,s)-ps*DB0D(0,m1s,s)));
}

inline Double DM1DeltaB0OnTwoPiI(Double s, Double m1se, Double m2se)
{ return((m1se-m2se-s)/(s*sla(s,m1se,m2se))); }

static Cplx T2234sub0k(Double s)
{
  return(-DM1DeltaB0OnTwoPiI(s,m1s,m2s)
           *(B0D(ps,m3s,s)-B0D(0,m3s,s)+ps/(2*s)));
}

static Cplx T1234C3sub0kp(Double s)
{
  return(-DeltaB0OnTwoPiI(s,m3s,m4s)*(B0D(ps,m2s,s)-B0D(0,m2s,s)+ps/(2*s)) );
}

static Cplx T11234C3M1Zsub0k(Double s)
{
  return((-DeltaB0OnTwoPiI(s,m3s,m4s)*(B0D(ps,m2s,s)-B0D(0,m2s,s)+ps/(2*s)))/
	 sqr(s));
}

static Cplx T11234fin0kp2(Double s)
{
  if (fabs((s-m1s)/(s+m1s))<T11234ASYMPLIMIT1)
    return(DeltaB0OnTwoPiI(s,m3s,m4s)*DM1B0(ps,(s+m1s)/2,m2s)*(m1s-s));
  else
    return(DeltaB0OnTwoPiI(s,m3s,m4s)
             *(B0A(ps,m2s,m1s)-B0A(ps,m2s,s) ));
}

static Cplx T11234fin0kDp2(Double s)
{
  return(DDeltaB0OnTwoPiI(s,m3s,m4s)*(B0A(ps,m1s,m2s)-B0A(ps,m2s,s))
         -DeltaB0OnTwoPiI(s,m3s,m4s)*DM1B0(ps,s,m2s));
}

Cplx T3aint(Double pse, Double m1se, Double m2se,
            Double m3se, Double prec, Double *abserr)
{
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se;
  threshold=sqr(sqrt(m2s)+sqrt(m3s));
  return(Cqaginf(T234sub0k,threshold,prec,abserr));
}

Cplx T3a1int(Double pse, Double m1se, Double m2se,
             Double m3se, Double prec, Double *abserr)
{
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se;
  threshold=sqr(sqrt(m1s)+sqrt(m2s));
  return(Cqaginf(T2234sub0k,threshold,prec,abserr));
}

Cplx T4aint(Double pse, Double m1se, Double m2se,
            Double m3se, Double m4se, Double prec, Double *abserr)
{
// following two lines added 09 Nov 99:
  if ((pse==m1se)&&(m2se==Double(0))&&(m3se==Double(0))&&(m4se==Double(0)))
    return Cplx(1);
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se; m4s=m4se;
  threshold=sqr(sqrt(m3s)+sqrt(m4s));
  return(Cqagpinf(T1234C3sub0kp,threshold,m1s,1,prec,abserr));
}

Cplx T11234C3M1Zsub0I(Double pse, Double m2se,
                      Double m3se, Double m4se, Double prec, Double *abserr)
{
  ps=pse; m2s=m2se; m3s=m3se; m4s=m4se;
  threshold=sqr(sqrt(m3s)+sqrt(m4s));
  return(Cqaginf(T11234C3M1Zsub0k,threshold,prec,abserr));
}

Cplx T4a1int(Double pse, Double m1se, Double m2se,
             Double m3se, Double m4se, Double prec, Double *abserr)
{
// following two lines added 09 Nov 99:
  if ((pse==m1se)&&(m2se==Double(0))&&(m3se==Double(0))&&(m4se==Double(0)))
    return (PiI-PiSonSix)/m1se;
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se; m4s=m4se;
  threshold=sqr(sqrt(m3s)+sqrt(m4s));
  return(Cqagp2inf(T11234fin0kp2,T11234fin0kDp2,threshold,m1s,1,prec,abserr));
}
