/* ------------------------------------------------------------- */
/* U6.cc                                                         */
/* Ayres Freitas (afreitas@pitt.edu)			         */
/* last revision 06.12.16                                        */
/* ------------------------------------------------------------- */
/* 3-loop master integral U6 with six                            */
/* independent propagators (subtracted finite part)    	         */
/* ------------------------------------------------------------- */
/* Reference:                                                    */
/*   [1] = Freitas, arXiv:1609.09159, JHEP 1611 (2016) 145       */
/* See also:                                                     */
/*   [2] = Bauberger, Boehm,                                     */
/*         hep-ph/9501201, Nucl.Phys.B 445 (1995) 25             */
/* ------------------------------------------------------------- */

#define Errstream cerr

#include "U456.h"
#include "li.h"
#include "qagutil.h"
#include "B0.h"
#include "C0.h"

/******************************************************************/
/* Throughout this file the numbers are defined as Double or Cplx   */
/* to enable a generalization to arbitrary precision              */
/******************************************************************/

/******************************************************************/
/* convention: vars is frequently used for (var)^2                */
/* for example: m1s denotes m_1^2.                                */
/******************************************************************/

/******************************************************************/
/* routines which cry for improvement are marked with             */
/* "TO IMPROVE".                                                  */
/******************************************************************/

/******************************************************************/
/* some constants which define from which value on the asymptotic */
/* limit of some functions will be taken:                         */
/******************************************************************/

#ifdef QUADPRECISION
static doubledouble ASYMP_LIMIT_DC01=SDouble("1e13");
static doubledouble ASYMP_LIMIT_T3K3_1=SDouble("1e-17");
#else
#define ASYMP_LIMIT_DC01 1e3
#define ASYMP_LIMIT_T3K3_1 1e-7
#endif


static Double ps,m1s,m2s,m3s,m4s,m5s,mxs;
static Double n1s,n2s,n3s,n4s,n5s,nxs;
static Double prec,abserr;
static Double threshold;
static Cplx B0fix,B0sfix;
static Cplx T2a(void);
static Cplx T2b(void);
static Cplx T3a(void);
static Cplx T3b(void);
static Cplx T2aC0(void);
static Cplx T2bC0(void);
static void increase_error(Double err);
static Cplx T2ak2(Double t);  /* integration kernel for T2a  */
static Cplx T2bk2(Double t);  /* integration kernel for T2b  */

// routines about the discontinuity of the one-loop triangle diagram,
// the mathematical
// background of the according algorithms is described in ref. [2]

#define p1s C0_p1s
#define p2s C0_p2s
#define p3s C0_p3s
#define m1s C0_m1s
#define m2s C0_m2s
#define m3s C0_m3s
#define m1 C0_m1
#define m2 C0_m2
#define m3 C0_m3
#define s0 C0_s0
#define a1 C0_a1
#define a2 C0_a2
#define Deltaps C0_Deltaps
#define Sumps C0_Sumps
#define status C0_status
#define s1_is_real C0_s1_is_real
#define Rs1 C0_Rs1
#define Rs2 C0_Rs2
#define Cs1 C0_Cs1

static Double p1s,p2s,p3s,m1s,m2s,m3s;
static Double m1,m2,m3,s0;
static Double a1,a2;  // for the fast calculation of a
static Double Deltaps,Sumps;
static int status;
static int C0_calc_status(void);
static int s1_is_real;
static Double Rs1,Rs2;
static Cplx Cs1;
static void C0dispersion_init(Double p2se,Double p3se,
			      Double m1se,Double m2se,Double m3se);
static Cplx C0_discontinuity(Double s);
static Cplx C0_discontinuitythr(void);


/*********************************************************************/
/* all what is needed to calculate the discontinuity of the one-loop */
/* three-point function C0, according to ref. [2]:                   */
/*********************************************************************/

static void C0dispersion_init(
    Double p2se,Double p3se,Double m1se,Double m2se,Double m3se)
{
  p2s=p2se; p3s=p3se; m1s=m1se; m2s=m2se; m3s=m3se;
  m1=sqrt(m1s); m2=sqrt(m2s); m3=sqrt(m3s);
  s0=sqr(m1+m2);
  a1=-p2s-p3s+m3s+m3s-m1s-m2s;
  a2=(m1s-m2s)*(p3s-p2s);
  Deltaps=sqr(sqrt(p2s)-sqrt(p3s));
  Sumps=sqr(sqrt(p2s)+sqrt(p3s));
  status=C0_calc_status();
}

// C0_calc_status returns the case the mass configuration belongs to    */
// according to [2]                                                  */

static int C0_calc_status(void)
{
  Double m3sthr;  // threshold value for m3s for the occurrence
                // of the anomalous threshold
  if (m1+m2==Double(0)) m3sthr=m3s+1;   // in fact m3sthr -> Infinity
  else m3sthr=(m1*p2s+m2*p3s)/(m1+m2)-m1*m2;
  if (m3s==Double(0)) {
    if (m3sthr<0) return(11);
    else return(12); };
  Double dm13s,sm13s,dm23s,sm23s;
  dm13s=sqr(m1-m3); sm13s=sqr(m1+m3); dm23s=sqr(m2-m3); sm23s=sqr(m2+m3);
  Cplx l3,l4;
  int l3_is_real,l4_is_real;
  if (p2s>sm23s) {l3_is_real=TRUE; l3=sla(p2s,m2s,m3s);}
  else if (p2s>dm23s) {l3_is_real=FALSE; l3=I*snla(p2s,m2s,m3s);}
       else {l3_is_real=TRUE; l3=-sla(p2s,m2s,m3s);};
  if (p3s>sm13s) {l4_is_real=TRUE; l4=sla(p3s,m1s,m3s);}
  else if (p3s>dm13s) {l4_is_real=FALSE; l4=I*snla(p3s,m1s,m3s);}
       else {l4_is_real=TRUE; l4=-sla(p3s,m1s,m3s);};
  if (l3_is_real==l4_is_real)
  {
    s1_is_real=TRUE;
    Rs1=(-m3s+p2s+p3s+m1s+m2s)*.5-((p2s-m2s)*(p3s-m1s)+real(l3*l4))*0.5/m3s;
    Rs2=(-m3s+p2s+p3s+m1s+m2s)*.5-((p2s-m2s)*(p3s-m1s)-real(l3*l4))*0.5/m3s;
  }
  else
  {
    s1_is_real=FALSE;
    Cs1=(-m3s+p2s+p3s+m1s+m2s)*0.5-((p2s-m2s)*(p3s-m1s)+l3*l4)*0.5/m3s;
  };
  if (m3s >= m3sthr) {return(1);}
  if (la(p2s,m2s,m3sthr)<0)
  {
    if (s1_is_real==TRUE)
    {
      if (Rs1<s0) return(2);
      else return(4);
    }
    else
    {
      return(3);
    };
  }
  else
  {
    if (s1_is_real==TRUE) return(5);
    else return(6);
  };
  Errstream << "Unknown status for the triangle discontinuity.\n";
  return(0);
}

// C0dispersion::discontinuity returns after a proper initialization */
// with p2s,p3s,m1s,m2s,m3s the discontinuity of the triangle        */
// diagram. It is assumed that s=p1s>(sqrt(m1s)+sqrt(m2s))^2

// TO IMPROVE:
//   there are many possibilities to improve the numerics in this routine
//   mainly: only one logarithm would be needed for each evaluation
static Cplx C0_discontinuity(Double s)
{
  Double a,b,l1,l2;
  Cplx Cb,Cl1;
  int b_is_complex; // indicates whether b is complex
  a=s*(s+a1)+a2;
  l2=sla(s,m1s,m2s);
  if (s<Deltaps) {b_is_complex=FALSE; l1=sla(s,p2s,p3s); b=l1*l2;}
  else if (s<Sumps) {b_is_complex=TRUE; Cl1=I*snla(s,p2s,p3s); Cb=Cl1*l2;}
  else  {b_is_complex=FALSE; l1=-sla(s,p2s,p3s); b=l1*l2;};
//  cout << status << " " << s << " " << l1 << " " << (l1==Double(0)) << " " << -TwoPiI*2*l2/a << endl;
  if ((status==1 || status==2 || status==4 || status==5) && 
      b_is_complex==FALSE && l1==Double(0))
    return(-TwoPiI*2*l2/a);
  if (status==1) {
#if DEBUGMASTER2
  cerr << "\nC0d: " << b_is_complex << " " << l1
       << " " << a+b << " " << a-b << "\n";
#endif
    if (b_is_complex==FALSE)
//      return(-(TwoPiI/l1)*(clog2(a+b)-clog2(a-b)));  
      return(-(TwoPiI/l1)*(clog2(a+b)-clog1(a-b)));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb))); };
  if (status==11) {
    if (b_is_complex==FALSE)
      if (s > ASYMP_LIMIT_DC01*(p2s+p3s+m1s+m2s)) {
                       // asymptotic formulae for large s:
        if ((p2s==m2s)&&(p3s!=m1s))
          return(-(TwoPiI/l1)*(clog1(2*p2s*(m1s-p3s)/s)-clog2(a-b)));
        if ((p2s!=m2s)&&(p3s==m1s))
          return(-(TwoPiI/l1)*(clog1(2*p3s*(m2s-p2s)/s)-clog2(a-b)));
        return(-(TwoPiI/l1)*(clog2(2.*(m1s-p3s)*(m2s-p2s))-clog2(a-b))); }
      else
        return(-(TwoPiI/l1)*(clog2(a+b)-clog2(a-b)));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb))); };
  if (status==12) {   // like status==4, but with s1=+infinity
    Cplx z1=-TwoPiI;
    if (b_is_complex==FALSE)
      if (s > ASYMP_LIMIT_DC01*(p2s+p3s+m1s+m2s)) { // asymptotic formulae:
        if ((p2s==m2s)&&(p3s!=m1s))
          return(-(TwoPiI/l1)*(clog1(2*p2s*(m1s-p3s)/s)-clog2(a-b)+z1));
        if ((p2s!=m2s)&&(p3s==m1s))
          return(-(TwoPiI/l1)*(clog1(2*p3s*(m2s-p2s)/s)-clog2(a-b)+z1));
        return(-(TwoPiI/l1)*(clog1(2.*(m1s-p3s)*(m2s-p2s))-clog2(a-b)+z1)); }
      else
        return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)+z1));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)+z1)); };
  if ((status==2)||(status==3)||(status==6)) {
    if (b_is_complex==FALSE)
      return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb))); };
  if (status==5) {
    if (s<Rs1) {
      if (b_is_complex==FALSE) {
        return(-(TwoPiI/l1)*(clog2(a+b)-clog2(a-b))); }
      else
        return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)-TwoPiI)); }
    else {
      if (b_is_complex==FALSE)
        return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)));
      else
        return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)));}; }
  if (status==4) {
    Cplx z1;
    if (s<Rs1) z1=-TwoPiI; else z1=Cplx(0);
    if (b_is_complex==FALSE)
      return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)+z1));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)+z1)); };
  return(Cplx(0));
}

static Cplx C0_discontinuitythr(void)   // at s=threshold
{
  Double a;
  Cplx Cl1,z1,z2;
  int l1_is_complex; // indicates whether l1 is complex
  Double s=sqr(sqrt(m1s)+sqrt(m2s));
  a=s*(s+a1)+a2;
  if (s<Deltaps) {l1_is_complex=FALSE; Cl1=sla(s,p2s,p3s); }
  else if (s<Sumps) {l1_is_complex=TRUE; Cl1=I*snla(s,p2s,p3s);}
  else  {l1_is_complex=FALSE; Cl1=-sla(s,p2s,p3s);};
  if ((status==1)||(status==11)) return(Cplx(0));
  if (status==12) {   // like status==4, but with s1=+infinity
    if (a<0) return(Cplx(0));
    else return(sqr(TwoPiI)/Cl1); };
  if ((status==2)||(status==3)||(status==6))
    if (a<0) return(-sqr(TwoPiI)/Cl1);
    else return(Cplx(0));
  if (status==5) {
    if (a<0) z2=TwoPiI; else z2=Cplx(0);
    if (s<Rs1) {
      if (l1_is_complex==FALSE) return(Cplx(0));
      else return(-TwoPiI*(z2-TwoPiI)/Cl1); }
    else {
      if (a<0) return(-sqr(TwoPiI)/Cl1); else return(Cplx(0)); } };
  if (status==4) {
    if (s<Rs1) z1=-TwoPiI; else z1=Cplx(0);
    if (a<0) z2=TwoPiI; else z2=Cplx(0);
    return(-(TwoPiI/Cl1)*(z2+z1)); };
  return(Cplx(0));
}


#undef p1s
#undef p2s
#undef p3s
#undef m1s
#undef m2s
#undef m3s
#undef m1
#undef m2
#undef m3
#undef s0
#undef a1
#undef a2
#undef Deltaps
#undef Sumps
#undef status
#undef s1_is_real
#undef Rs1
#undef Rs2
#undef Cs1

// routines about the discontinuity of the one-loop triangle diagram,
// the mathematical
// background of the according algorithms is described in ref. [2]

#define p1s C0s_p1s
#define p2s C0s_p2s
#define p3s C0s_p3s
#define m1s C0s_m1s
#define m2s C0s_m2s
#define m3s C0s_m3s
#define m1 C0s_m1
#define m2 C0s_m2
#define m3 C0s_m3
#define s0 C0s_s0
#define a1 C0s_a1
#define a2 C0s_a2
#define Deltaps C0s_Deltaps
#define Sumps C0s_Sumps
#define status C0s_status
#define s1_is_real C0s_s1_is_real
#define Rs1 C0s_Rs1
#define Rs2 C0s_Rs2
#define Cs1 C0s_Cs1

static Double p1s,p2s,p3s,m1s,m2s,m3s;
static Double m1,m2,m3,s0;
static Double a1,a2;  // for the fast calculation of a
static Double Deltaps,Sumps;
static int status;
static int C0s_calc_status(void);
static int s1_is_real;
static Double Rs1,Rs2;
static Cplx Cs1;
static void C0sdispersion_init(Double p2se,Double p3se,
			      Double m1se,Double m2se,Double m3se);
static Cplx C0s_discontinuity(Double s);
static Cplx C0s_discontinuitythr(void);


/*********************************************************************/
/* all what is needed to calculate the discontinuity of the one-loop */
/* three-point function C0, according to ref. [2]:                   */
/*********************************************************************/

static void C0sdispersion_init(
    Double p2se,Double p3se,Double m1se,Double m2se,Double m3se)
{
  p2s=p2se; p3s=p3se; m1s=m1se; m2s=m2se; m3s=m3se;
  m1=sqrt(m1s); m2=sqrt(m2s); m3=sqrt(m3s);
  s0=sqr(m1+m2);
  a1=-p2s-p3s+m3s+m3s-m1s-m2s;
  a2=(m1s-m2s)*(p3s-p2s);
  Deltaps=sqr(sqrt(p2s)-sqrt(p3s));
  Sumps=sqr(sqrt(p2s)+sqrt(p3s));
  status=C0s_calc_status();
}

// C0_calc_status returns the case the mass configuration belongs to    */
// according to [2]                                                  */

static int C0s_calc_status(void)
{
  Double m3sthr;  // threshold value for m3s for the occurrence
                // of the anomalous threshold
  if (m1+m2==Double(0)) m3sthr=m3s+1;   // in fact m3sthr -> Infinity
  else m3sthr=(m1*p2s+m2*p3s)/(m1+m2)-m1*m2;
  if (m3s==Double(0)) {
    if (m3sthr<0) return(11);
    else return(12); };
  Double dm13s,sm13s,dm23s,sm23s;
  dm13s=sqr(m1-m3); sm13s=sqr(m1+m3); dm23s=sqr(m2-m3); sm23s=sqr(m2+m3);
  Cplx l3,l4;
  int l3_is_real,l4_is_real;
  if (p2s>sm23s) {l3_is_real=TRUE; l3=sla(p2s,m2s,m3s);}
  else if (p2s>dm23s) {l3_is_real=FALSE; l3=I*snla(p2s,m2s,m3s);}
       else {l3_is_real=TRUE; l3=-sla(p2s,m2s,m3s);};
  if (p3s>sm13s) {l4_is_real=TRUE; l4=sla(p3s,m1s,m3s);}
  else if (p3s>dm13s) {l4_is_real=FALSE; l4=I*snla(p3s,m1s,m3s);}
       else {l4_is_real=TRUE; l4=-sla(p3s,m1s,m3s);};
  if (l3_is_real==l4_is_real)
  {
    s1_is_real=TRUE;
    Rs1=(-m3s+p2s+p3s+m1s+m2s)*.5-((p2s-m2s)*(p3s-m1s)+real(l3*l4))*0.5/m3s;
    Rs2=(-m3s+p2s+p3s+m1s+m2s)*.5-((p2s-m2s)*(p3s-m1s)-real(l3*l4))*0.5/m3s;
  }
  else
  {
    s1_is_real=FALSE;
    Cs1=(-m3s+p2s+p3s+m1s+m2s)*0.5-((p2s-m2s)*(p3s-m1s)+l3*l4)*0.5/m3s;
  };
  if (m3s >= m3sthr) {return(1);}
  if (la(p2s,m2s,m3sthr)<0)
  {
    if (s1_is_real==TRUE)
    {
      if (Rs1<s0) return(2);
      else return(4);
    }
    else
    {
      return(3);
    };
  }
  else
  {
    if (s1_is_real==TRUE) return(5);
    else return(6);
  };
  Errstream << "Unknown status for the triangle discontinuity.\n";
  return(0);
}

// C0dispersion::discontinuity returns after a proper initialization */
// with p2s,p3s,m1s,m2s,m3s the discontinuity of the triangle        */
// diagram. It is assumed that s=p1s>(sqrt(m1s)+sqrt(m2s))^2

// TO IMPROVE:
//   there are many possibilities to improve the numerics in this routine
//   mainly: only one logarithm would be needed for each evaluation
static Cplx C0s_discontinuity(Double s)
{
  Double a,b,l1,l2;
  Cplx Cb,Cl1;
  int b_is_complex; // indicates whether b is complex
  a=s*(s+a1)+a2;
  l2=sla(s,m1s,m2s);
  if (s<Deltaps) {b_is_complex=FALSE; l1=sla(s,p2s,p3s); b=l1*l2;}
  else if (s<Sumps) {b_is_complex=TRUE; Cl1=I*snla(s,p2s,p3s); Cb=Cl1*l2;}
  else  {b_is_complex=FALSE; l1=-sla(s,p2s,p3s); b=l1*l2;};
  if (status==1) {
#if DEBUGMASTER2
  cerr << "\nC0d: " << b_is_complex << " " << l1
       << " " << a+b << " " << a-b << "\n";
#endif
    if (b_is_complex==FALSE)
//      return(-(TwoPiI/l1)*(clog2(a+b)-clog2(a-b)));  
      return(-(TwoPiI/l1)*(clog2(a+b)-clog1(a-b)));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb))); };
  if (status==11) {
    if (b_is_complex==FALSE)
      if (s > ASYMP_LIMIT_DC01*(p2s+p3s+m1s+m2s)) {
                       // asymptotic formulae for large s:
        if ((p2s==m2s)&&(p3s!=m1s))
          return(-(TwoPiI/l1)*(clog1(2*p2s*(m1s-p3s)/s)-clog2(a-b)));
        if ((p2s!=m2s)&&(p3s==m1s))
          return(-(TwoPiI/l1)*(clog1(2*p3s*(m2s-p2s)/s)-clog2(a-b)));
        return(-(TwoPiI/l1)*(clog2(2.*(m1s-p3s)*(m2s-p2s))-clog2(a-b))); }
      else
        return(-(TwoPiI/l1)*(clog2(a+b)-clog2(a-b)));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb))); };
  if (status==12) {   // like status==4, but with s1=+infinity
    Cplx z1=-TwoPiI;
    if (b_is_complex==FALSE)
      if (s > ASYMP_LIMIT_DC01*(p2s+p3s+m1s+m2s)) { // asymptotic formulae:
        if ((p2s==m2s)&&(p3s!=m1s))
          return(-(TwoPiI/l1)*(clog1(2*p2s*(m1s-p3s)/s)-clog2(a-b)+z1));
        if ((p2s!=m2s)&&(p3s==m1s))
          return(-(TwoPiI/l1)*(clog1(2*p3s*(m2s-p2s)/s)-clog2(a-b)+z1));
        return(-(TwoPiI/l1)*(clog1(2.*(m1s-p3s)*(m2s-p2s))-clog2(a-b)+z1)); }
      else
        return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)+z1));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)+z1)); };
  if ((status==2)||(status==3)||(status==6)) {
    if (b_is_complex==FALSE)
      return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb))); };
  if (status==5) {
    if (s<Rs1) {
      if (b_is_complex==FALSE) {
        return(-(TwoPiI/l1)*(clog2(a+b)-clog2(a-b))); }
      else
        return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)-TwoPiI)); }
    else {
      if (b_is_complex==FALSE)
        return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)));
      else
        return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)));}; }
  if (status==4) {
    Cplx z1;
    if (s<Rs1) z1=-TwoPiI; else z1=Cplx(0);
    if (b_is_complex==FALSE)
      return(-(TwoPiI/l1)*(clog1(a+b)-clog2(a-b)+z1));
    else
      return(-(TwoPiI/Cl1)*(log(a+Cb)-log(a-Cb)+z1)); };
  return(Cplx(0));
}

static Cplx C0s_discontinuitythr(void)   // at s=threshold
{
  Double a;
  Cplx Cl1,z1,z2;
  int l1_is_complex; // indicates whether l1 is complex
  Double s=sqr(sqrt(m1s)+sqrt(m2s));
  a=s*(s+a1)+a2;
  if (s<Deltaps) {l1_is_complex=FALSE; Cl1=sla(s,p2s,p3s); }
  else if (s<Sumps) {l1_is_complex=TRUE; Cl1=I*snla(s,p2s,p3s);}
  else  {l1_is_complex=FALSE; Cl1=-sla(s,p2s,p3s);};
  if ((status==1)||(status==11)) return(Cplx(0));
  if (status==12) {   // like status==4, but with s1=+infinity
    if (a<0) return(Cplx(0));
    else return(sqr(TwoPiI)/Cl1); };
  if ((status==2)||(status==3)||(status==6))
    if (a<0) return(-sqr(TwoPiI)/Cl1);
    else return(Cplx(0));
  if (status==5) {
    if (a<0) z2=TwoPiI; else z2=Cplx(0);
    if (s<Rs1) {
      if (l1_is_complex==FALSE) return(Cplx(0));
      else return(-TwoPiI*(z2-TwoPiI)/Cl1); }
    else {
      if (a<0) return(-sqr(TwoPiI)/Cl1); else return(Cplx(0)); } };
  if (status==4) {
    if (s<Rs1) z1=-TwoPiI; else z1=Cplx(0);
    if (a<0) z2=TwoPiI; else z2=Cplx(0);
    return(-(TwoPiI/Cl1)*(z2+z1)); };
  return(Cplx(0));
}


#undef p1s
#undef p2s
#undef p3s
#undef m1s
#undef m2s
#undef m3s
#undef m1
#undef m2
#undef m3
#undef s0
#undef a1
#undef a2
#undef Deltaps
#undef Sumps
#undef status
#undef s1_is_real
#undef Rs1
#undef Rs2
#undef Cs1


/*********************************************************************/
/* The two-loop master integral,                                     */
/* according to ref. [2] and [5]:                                    */
/*********************************************************************/

inline bool BelowThreshold2(Double ps,Double m1s,Double m2s)
{ return(ps<sqr(sqrt(m1s)+sqrt(m2s))); }

inline bool BelowThreshold3(Double ps,Double m1s,Double m2s,Double m3s)
{ return(ps<sqr(sqrt(m1s)+sqrt(m2s)+sqrt(m3s))); }


inline Double tr2(Double x)
{
  return(x/((1-x)*(1-x)));
}

inline Double Dtr2(Double x)   // the derivative of tr2
{
  return((1+x)/((1-x)*(1-x)*(1-x)));
}

inline Double tr2inv(Double y)  // the inverse of tr2
{
  if (y==Double(0)) return(0);
  return(.5*(1.+2.*y-sqrt(1+4.*y))/y);
}

static Cplx Tsk(Double x)
{
  Double err;
  
  ps = tr2(x);
  Double tt = real(T2aC0()+T2bC0()+T3a()+T3b())*Dtr2(x);
  return(tt);
}

Double U6int(Double m1se,Double m2se,Double m3se,Double m4se,Double m5se,
	     Double m6se,Double prece,Double *abserre)
{
  m1s=m1se; m2s=m2se; m3s=m6se; m4s=m4se; m5s=m3se; mxs=m5se; prec=prece;
  n1s=m6se; n2s=m6se; n3s=m6se; n4s=m6se; n5s=m6se; nxs=m6se;
  Double err;
//  cout << prec*100 << endl;
  Double lim1 = tr2inv(0), lim2 = tr2inv(SDouble("1e10")*m6se),
          lim3 = tr2inv(SDouble("2e10")*m6se);
  Cplx z=Cqag(Tsk,lim1,lim2,prec*100,&err);
// extrapolate from lim2 to 1 using parametrization (1-x)*(a+b*log(1-x)) :
  Cplx y2 = Tsk(lim2), y3 = Tsk(lim3);
  Double det = (1-lim2)*(1-lim3)*log((1-lim3)/(1-lim2));
  Cplx a = -((1-lim2)*log(1-lim2)*y3 - (1-lim3)*log(1-lim3)*y2)/det,
       b = ((1-lim2)*y3 - (1-lim3)*y2)/det;
  z += sqr(1-lim2)/4 * (2*a - b + 2*b*log(1-lim2));
  *abserre = err;
  return(real(z));
}


static Cplx T2ak2(Double t)
{
  Cplx tres = Double(0);
  if(t > C0_s0)  tres += conj(C0_discontinuity(t)/TwoPiI)*B0fix;
  if(t > C0s_s0)  tres -= conj(C0s_discontinuity(t)/TwoPiI)*B0sfix;
  return(tres);
}

static Cplx T2bk2(Double t)
{
  Cplx tres = Double(0);
  if(t > C0_s0)  tres += C0_discontinuity(t)/TwoPiI*conj(B0fix);
  if(t > C0s_s0)  tres -= C0s_discontinuity(t)/TwoPiI*conj(B0sfix);
  return(tres);
}

static Cplx T3ak2(Double t)
{
  return(real(conj(C0_discontinuity(t)/TwoPiI)*DeltaB0OnTwoPiI(ps,t,m2s)*B0fix));
}

static Cplx T3ak2s(Double t)
{
  return(real(conj(C0_discontinuity(t)/TwoPiI)*DeltaB0OnTwoPiI(ps,t,n2s)*B0fix));
}

static Cplx T4ak2(Double t)
{
  return(real(conj(C0_discontinuity(t)/TwoPiI)*DeltaB0OnTwoPiI(ps,t,m1s)*B0fix));
}

static Cplx T4ak2s(Double t)
{
  return(real(conj(C0_discontinuity(t)/TwoPiI)*DeltaB0OnTwoPiI(ps,t,n1s)*B0fix));
}


// the two-particle contributions T2a and T2b (see [2]) are very
// similar. It simplifies therefore the code developpment to define the masses
// with aliases:

#define M1S m1s
#define M2S m2s
#define M4S m4s
#define M5S m5s
#define N1S n1s
#define N2S n2s
#define N4S n4s
#define N5S n5s
#define TM22 T2ak2

static Cplx T2a(void)
{
  C0dispersion_init(M1S,M2S,M5S,M4S,m3s);
  C0sdispersion_init(N1S,N2S,N5S,N4S,n3s);
  Cplx z1,z2;
  Double threshold1=sqr(sqrt(M4S)+sqrt(M5S)), 
         thresholds=sqr(sqrt(N4S)+sqrt(N5S));
  Double err1=0,err2=0;
  Double M1=sqrt(M1S),M2=sqrt(M2S);
  Double lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  if(ps > lim2)
    B0fix = -(B0A(0,ps,mxs)+2)*DeltaB0OnTwoPiI(ps,M1S,M2S);
  else
    B0fix = Double(0);
  if(ps > sqr(sqrt(N1S)+sqrt(N2S)))
    B0sfix= -(B0A(0,ps,nxs)+2)*DeltaB0OnTwoPiI(ps,N1S,N2S);
  else
    B0sfix= Double(0);
  if(thresholds < threshold1)
    threshold1 = thresholds;
//  cout << "> " << TM22(25.)/(25.-ps) << " " << C0_discontinuity(25.) << endl;
  z1=Cqagpinf(TM22,threshold1,ps,1,prec,&err1,lim1,lim2);
//  z1 = Cqaginf(TM21,threshold1,prec,&err1);
  increase_error(abs(err1));
  return(z1+z2/TwoPiI);
}

static Cplx T2aC0(void)
{
  Double M1=sqrt(M1S),M2=sqrt(M2S);
  Double lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  if(ps > lim2)
    B0fix = -(B0A(0,ps,mxs)+2)*DeltaB0OnTwoPiI(ps,M1S,M2S);
  else
    B0fix = Double(0);
  if(ps > sqr(sqrt(N1S)+sqrt(N2S)))
    B0sfix= -(B0A(0,ps,nxs)+2)*DeltaB0OnTwoPiI(ps,N1S,N2S);
  else
    B0sfix= Double(0);
  return(B0fix*conj(C0Q(ps,M1S,M2S,M5S,M4S,m3s)) - 
         B0sfix*conj(C0Q(ps,N1S,N2S,N5S,N4S,n3s)));
}


#undef M1S
#undef M2S
#undef M4S
#undef M5S
#undef N1S
#undef N2S
#undef N4S
#undef N5S
#undef TM21
#undef TM22
#undef TM23
#undef TM24


#define M1S m4s
#define M2S m5s
#define M4S m1s
#define M5S m2s
#define N1S n4s
#define N2S n5s
#define N4S n1s
#define N5S n2s
#define TM22 T2bk2

static Cplx T2b(void)
{
  C0dispersion_init(M1S,M2S,M5S,M4S,m3s);
  C0sdispersion_init(N1S,N2S,N5S,N4S,n3s);
  Cplx z1,z2;
  Double threshold1=sqr(sqrt(M4S)+sqrt(M5S)), 
         thresholds=sqr(sqrt(N4S)+sqrt(N5S));
  Double err1=0,err2=0;
  Double M1=sqrt(M1S),M2=sqrt(M2S);
  Double lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  if(ps > lim2)
    B0fix = -(B0A(0,ps,mxs)+2)*DeltaB0OnTwoPiI(ps,M1S,M2S);
  else
    B0fix = Double(0);
  if(ps > sqr(sqrt(N1S)+sqrt(N2S)))
    B0sfix= -(B0A(0,ps,nxs)+2)*DeltaB0OnTwoPiI(ps,N1S,N2S);
  else
    B0sfix= Double(0);
  if(thresholds < threshold1)
    threshold1 = thresholds;
  z1=Cqagpinf(TM22,threshold1,ps,1,prec,&err1,lim1,lim2);
//  z1 = Cqaginf(TM21,threshold1,prec,&err1);
  increase_error(abs(err1));
  return(z1+z2/TwoPiI);
}

static Cplx T2bC0(void)
{
  Double M1=sqrt(M1S),M2=sqrt(M2S);
  Double lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  if(ps > lim2)
    B0fix = -(B0A(0,ps,mxs)+2)*DeltaB0OnTwoPiI(ps,M1S,M2S);
  else
    B0fix = Double(0);
  if(ps > sqr(sqrt(N1S)+sqrt(N2S)))
    B0sfix= -(B0A(0,ps,nxs)+2)*DeltaB0OnTwoPiI(ps,N1S,N2S);
  else
    B0sfix= Double(0);
  return(B0fix*conj(C0Q(ps,M1S,M2S,M5S,M4S,m3s)) - 
         B0sfix*conj(C0Q(ps,N1S,N2S,N5S,N4S,n3s)));
}

#undef M1S
#undef M2S
#undef M4S
#undef M5S
#undef N1S
#undef N2S
#undef N4S
#undef N5S
#undef TM21
#undef TM22
#undef TM23
#undef TM24

Cplx Cqagp1(Cplx (*f) (Double),Double a,Double b,Double singe,int iepssign,
                   Double epsrel,Double *error);

static Cplx T3a(void)
{
  Cplx z1,z2,z1s;
  C0dispersion_init(m2s,ps,m4s,m3s,m5s);
  Double threshold1=sqr(sqrt(m3s)+sqrt(m4s)),
         threshold2=sqr(sqrt(ps)-sqrt(m2s));
  Double err1=0,err2=0;
  Double M1=sqrt(m2s),M2=sqrt(ps);
  Double lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  B0fix = (B0A(0,ps,mxs)+2);
  if(ps<sqr(sqrt(m2s)+sqrt(m3s)+sqrt(m4s))) 
    z1 = Cplx(0);
  else
    z1 = Cqagp1(T3ak2,threshold1,threshold2,m1s,1,prec,&err1);
  increase_error(abs(err1));
  C0dispersion_init(n2s,ps,n4s,n3s,n5s);
  threshold1=sqr(sqrt(n3s)+sqrt(n4s));
  threshold2=sqr(sqrt(ps)-sqrt(n2s));
  err1=err2=0;
  M1=sqrt(n2s);
  lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  B0fix = (B0A(0,ps,nxs)+2);
  if(ps<sqr(sqrt(n2s)+sqrt(n3s)+sqrt(n4s))) 
    z1s = Cplx(0);
  else
    z1s = Cqagp1(T3ak2s,threshold1,threshold2,n1s,1,prec,&err1);
  return(z1-z1s/*+z2/TwoPiI*/);
}

static Cplx T3b(void)
{
  Cplx z1,z2,z1s;
  C0dispersion_init(m1s,ps,m5s,m3s,m4s);
  Double threshold1=sqr(sqrt(m3s)+sqrt(m5s)),
         threshold2=sqr(sqrt(ps)-sqrt(m1s));
  Double err1=0,err2=0;
  Double M1=sqrt(m1s),M2=sqrt(ps);
  Double lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  B0fix = (B0A(0,ps,mxs)+2);
  if(ps<sqr(sqrt(m1s)+sqrt(m3s)+sqrt(m5s))) 
    z1 = Cplx(0);
  else
    z1 = Cqagp1(T4ak2,threshold1,threshold2,m2s,1,prec,&err1);
  increase_error(abs(err1));
  C0dispersion_init(n1s,ps,n5s,n3s,n4s);
  threshold1=sqr(sqrt(n3s)+sqrt(n5s));
  threshold2=sqr(sqrt(ps)-sqrt(n1s));
  err1=err2=0;
  M1=sqrt(n1s);
  lim1=sqr(M1-M2),lim2=sqr(M1+M2);
  B0fix = (B0A(0,ps,nxs)+2);
  if(ps<sqr(sqrt(n1s)+sqrt(n3s)+sqrt(n5s))) 
    z1s = Cplx(0);
  else
    z1s = Cqagp1(T4ak2s,threshold1,threshold2,n2s,1,prec,&err1);
  return(z1-z1s/*+z2/TwoPiI*/);
}


static void increase_error(Double err)
{
  abserr=sqrt(sqr(err)+sqr(abserr));
}
