/* ------------------------------------------------------------- */
/* Uabc.cc                                                       */
/* Ayres Freitas (afreitas@pitt.edu)                             */
/* last revision 06.08.19                                        */
/* ------------------------------------------------------------- */
/* 3-loop self-energy master integrals where both sub-loops are  */
/* 2-propagator bubbles (subtracted finite parts)                */
/* ------------------------------------------------------------- */
/* Reference:                                                    */
/*   [2] = Bauberger, Freitas, Wiegand, arXiv:1908.09887         */
/* ------------------------------------------------------------- */

#include "Uabc.h"
#include "li.h"
#include "qagutil.h"
#include "B0.h"
#include "tinline.h"

#define Errstream cerr

/******************************************************************/
/* Throughout this file the numbers are defined as Double or Cplx */
/* to enable a generalization to arbitrary precision              */
/******************************************************************/

/******************************************************************/
/* convention: vars is frequently used for (var)^2                */
/* for example: m1s denotes m_1^2.                                */
/******************************************************************/


/****** Self-energy integrals with single-power propagators: ******/

static Double ps,m1s,m2s,m3s,m4s,m5s,m6s,s1;
static Double prec,abserr;
static Cplx U4akk(Double x);
static Cplx U4ak(Double x);
static Cplx U5ak1(Double x);
static Cplx U5ak2(Double x);
static Cplx U5bkk(Double x);
static Cplx U5bk(Double x);
static Cplx U5ckk(Double x);
static Cplx U5ck(Double x);
static Cplx U6ak1(Double x);
static Cplx U6ak2(Double x);
static Cplx U6bkk(Double x);
static Cplx U6bk(Double x);
static Cplx U6ckk(Double x);
static Cplx U6ck(Double x);


Double DB0z(Double m1s, Double m2s)
{
  if(abs(m1s-m2s) < 1e-5*m1s)
  {
    Double x = m2s/m1s-1;
    return((1 - x/2 + 3*x*x/10 - x*x*x/5 + powint(x,4)/7)/(6*m1s));
  } 
  else  
    return((m1s*m1s-m2s*m2s-2*m1s*m2s*log(m1s/m2s))/(2*powint(m1s-m2s,3)));
}

Double DDB0z(Double m1s, Double m2s)
{
  if(abs(m1s-m2s) < 1e-4*m1s)
  {
    Double x = m2s/m1s-1;
    return((1 - x + 6*x*x/7 - 5*x*x*x/7 + 25*powint(x,4)/42)/(30*m1s*m1s));
  } 
  else  
    return(((m1s-m2s)*(m1s*m1s+10*m1s*m2s+m2s*m2s) -
    	    6*m1s*m2s*(m1s+m2s)*log(m1s/m2s))/(3*powint(m1s-m2s,5)));
}

Double DDDB0z(Double m1s, Double m2s)
{
  if(abs(m1s-m2s) < 5e-3*m1s)
  {
    Double x = m2s/m1s-1;
    return((1 - 3*x/2 + 5*x*x/3 - 5*x*x*x/3 + 35*powint(x,4)/22 
            - 49*powint(x,5)/33)/(70*m1s*m1s*m1s));
  } 
  else  
    return(((m1s*m1s-m2s*m2s)*(m1s*m1s+28*m1s*m2s+m2s*m2s) -
    	    12*m1s*m2s*(m1s*m1s+3*m1s*m2s+m2s*m2s)*log(m1s/m2s))/
	   (2*powint(m1s-m2s,7)));
}

Double DDM1B0z(Double m2s, Double m1s)
{
  if(abs(m1s-m2s) < 1e-5*m1s)
  {
    Double x = m2s/m1s-1;
    return((-1 + 6*x/5 - 6*x*x/5 + 8*x*x*x/7)/(12*m1s*m1s));
  } 
  else  
    return(((m1s-m2s)*(5*m1s+m2s)-2*m1s*(m1s+2*m2s)*log(m1s/m2s))/
    	   (2*powint(m1s-m2s,4)));
}

Double DDDM1B0z(Double m1s, Double m2s)
{
  if(abs(m1s-m2s) < 1e-4*m1s)
  {
    Double x = m2s/m1s-1;
    return((-1 + 12*x/7 - 15*x*x/7 + 50*x*x*x/21)/(30*m1s*m1s*m1s));
  } 
  else  
    return((2*(m1s-m2s)*(10*m1s*m1s+19*m1s*m2s+m2s*m2s) -
    	    6*m1s*(m1s*m1s+6*m1s*m2s+3*m2s*m2s)*log(m1s/m2s))/
	   (3*powint(m1s-m2s,6)));
}


static Cplx U4akk(Double s2)
{
  Cplx z;
  if(s1+s2 > ps*2e5)
    z = DeltaB0(s2,m3s,m4s)*(sqr(ps)*ps/6*DDDB0z(s1,s2));
  else
    z = DeltaB0(s2,m3s,m4s)*(B0A(ps,s1,s2) - B0A(0,s1,s2) - ps*DB0z(s1,s2)
    				- sqr(ps)/2*DDB0z(s1,s2));
  return(z);
}

static Cplx U4ak(Double x)
{
  Double err1=0;
  s1=x;
  return(DeltaB0(s1,m1s,m2s) * 
  	 Cqaginf2(U4akk,sqr(sqrt(m3s)+sqrt(m4s)),prec,&err1));
}

Cplx U4aint(Double pse,Double m1se,Double m3se,Double m6se,
            Double m8se,Double prece,Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m3se; m3s=m6se; m4s=m8se; prec=prece;
  Double err;
  Cplx z=Cqaginf(U4ak,sqr(sqrt(m1s)+sqrt(m2s)),prec*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U5ak1(Double s)
{
  return(-DeltaB0(s,m3s,m4s)*real(B0A(s,m5s,m6s) - B0A(0,m5s,m6s))*
  	 (B0D(ps,m2s,s) - B0D(0,m2s,s) - ps*DB0D(0,m2s,s)));
}

static Cplx U5ak2(Double s)
{
  return(-DeltaB0(s,m5s,m6s)*real(B0A(s,m3s,m4s) - B0A(0,m3s,m4s))*
  	 (B0D(ps,m2s,s) - B0D(0,m2s,s) - ps*DB0D(0,m2s,s)));
}

Cplx U5aint(Double pse,Double m1se,Double m3se,Double m5se,
            Double m6se,Double m7se,Double prece,Double *abserre)
{
  ps=pse; m2s=m5se; m3s=m1se; m4s=m3se; m5s=m6se; m6s=m7se; prec=prece;
  Double err1, err2;
  Cplx z = Cqaginf(U5ak1,sqr(sqrt(m3s)+sqrt(m4s)),prec,&err1)
          +Cqaginf(U5ak2,sqr(sqrt(m5s)+sqrt(m6s)),prec,&err2);
  *abserre = sqrt(sqr(err1)+sqr(err2));
  return(z);
}


static Cplx U5bkk(Double s2)
{
  Cplx z;
  if(s1+s2 > ps*1e7)
    z = DeltaB0(s2,m5s,m6s)*(sqr(ps)/2*DDB0z(s1,s2));
  else
    z = DeltaB0(s2,m5s,m6s)*(B0A(ps,s1,s2) - B0A(0,s1,s2) - ps*DB0z(s1,s2));
  return(z);
}

static Cplx U5bk(Double x)
{
  Double err1=0;
  s1=x;
  return(DeltaB0(s1,m3s,m4s) * 
  	 Cqaginf2(U5bkk,sqr(sqrt(m5s)+sqrt(m6s)),prec,&err1));
}

Cplx U5bint(Double pse,Double m2se,Double m3se,Double m4se,
            Double m6se,Double m7se,Double prece,Double *abserre)
{
  ps=pse; m1s=m4se; m3s=m6se; m4s=m7se; m5s=m2se; m6s=m3se; prec=prece;
  Double err;
  Cplx z=Cqagpinf(U5bk,sqr(sqrt(m3s)+sqrt(m4s)),m1s,1,prec*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U5ckk(Double s2)
{
  Cplx z;
  z = DeltaB0(s2,m5s,s1)*(B0D(ps,m2s,s2) - B0D(0,m2s,s2) - ps*DB0D(0,m2s,s2));
  return(z);
}

static Cplx U5ck(Double x)
{
  Double err1=0;
  s1=x;
  return(DeltaB0(s1,m3s,m4s) * 
  	 Cqagpinf2(U5ckk,sqr(sqrt(m5s)+sqrt(s1)),m1s,1,prec,&err1));
}

Cplx U5cint(Double pse,Double m1se,Double m2se,Double m3se,Double m6se,
            Double m7se,Double prece,Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se; m4s=m6se; m5s=m7se; prec=prece;
  Double err;
  Cplx z=Cqaginf(U5ck,sqr(sqrt(m3s)+sqrt(m4s)),prec*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U6ak1(Double s)
{
  return(-DeltaB0(s,m3s,m4s)*real(B0A(s,m5s,m6s) - B0A(m1s,m5s,m6s))*
  	 (B0C(ps,m2s,s) - B0C(0,m2s,s)));
}

static Cplx U6ak2(Double s)
{
  return(-DeltaB0(s,m5s,m6s)*real(B0A(s,m3s,m4s) - B0A(m1s,m3s,m4s))*
  	 (B0C(ps,m2s,s) - B0C(0,m2s,s)));
}

Cplx U6aint(Double pse,Double m1se,Double m3se,Double m4se,Double m5se,
            Double m6se,Double m7se,Double prece,Double *abserre)
{
  ps=pse; m1s=m4se; m2s=m5se; m3s=m1se; m4s=m3se; m5s=m6se; m6s=m7se; prec=prece;
  Double err1, err2;
  Cplx z = Cqagpinf(U6ak1,sqr(sqrt(m3s)+sqrt(m4s)),m1s,1,prec,&err1)
          +Cqagpinf(U6ak2,sqr(sqrt(m5s)+sqrt(m6s)),m1s,1,prec,&err2);
  *abserre = sqrt(sqr(err1)+sqr(err2));
  return(z);
}


static Cplx U6bkk(Double s2)
{
  Cplx z;
  if(s1+s2 > ps*1e8)
    z = DeltaB0(s2,m5s,m6s)*(ps*DB0z(s1,s2));
  else
    z = DeltaB0(s2,m5s,m6s)*(B0A(ps,s1,s2) - B0A(0,s1,s2));
  return(z);
}

static Cplx U6bk(Double x)
{
  Double err1=0;
  s1=x;
  return(DeltaB0(s1,m3s,m4s) * 
  	 Cqagpinf2(U6bkk,sqr(sqrt(m5s)+sqrt(m6s)),m2s,1,prec,&err1));
}

Cplx U6bint(Double pse,Double m1se,Double m3se,Double m4se,Double m5se,
            Double m6se,Double m8se,Double prece,Double *abserre)
{
  ps=pse; m1s=m4se; m2s=m5se; m3s=m1se; m4s=m3se; m5s=m6se; m6s=m8se; prec=prece;
  Double err;
  Cplx z=Cqagpinf(U6bk,sqr(sqrt(m3s)+sqrt(m4s)),m1s,1,prec*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U6ckk(Double s2)
{
  Cplx z;
  z = DeltaB0(s2,m4s,s1)*(B0C(ps,m2s,s2) - B0C(0,m2s,s2));
  return(z);
}

static Cplx U6ck(Double x)
{
  Double err1=0;
  s1=x;
  return(DeltaB0(s1,m5s,m6s) * 
  	 Cqagpinf2(U6ckk,sqr(sqrt(m4s)+sqrt(s1)),m1s,1,prec,&err1));
}

Cplx U6cint(Double pse,Double m1se,Double m2se,Double m3se,Double m4se,
            Double m6se,Double m7se,Double prece,Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m2se; m3s=m4se; m4s=m3se; m5s=m6se; m6s=m7se; prec=prece;
  Double err;
  Cplx z=Cqagpinf(U6ck,sqr(sqrt(m5s)+sqrt(m6s)),m3s,1,prec*100,&err);
  *abserre = err;
  return(z);
}


/****** Self-energy integrals with double-power propagators: ******/

static Cplx U4a1kk(Double x);
static Cplx U4a1k(Double x);
static Cplx U4a2kk(Double x);
static Cplx U4a2k(Double x);
static Cplx U4a3kk(Double x);
static Cplx U4a3k(Double x);
static Cplx U4a3k1(Double x);
static Cplx U5a1k1(Double x);
static Cplx U5a1k2(Double x);
static Cplx U5a2k1(Double x);
static Cplx U5a2k2(Double x);
static Cplx U5b1kk(Double x);
static Cplx DU5b1kk(Double x);
static Cplx U5b1k(Double x);
static Cplx U5b2kk(Double x);
static Cplx U5b2k(Double x);
static Cplx U5c1kk(Double x);
static Cplx U5c1k(Double x);


static Cplx U4a1kk(Double s2)
{
  Cplx z;
  if(s1+s2 > ps*2e6)
    z = DM1DeltaB0(s2,m1s,m2s)*(sqr(ps)/2*DDB0z(s1,s2));
  else
    z = DM1DeltaB0(s2,m1s,m2s)*(B0A(ps,s1,s2) - B0A(0,s1,s2) - ps*DB0z(s1,s2));
  return(z);
}

static Cplx U4a1k(Double x)
{
  Double err1=0;
  s1=x;
  return(DeltaB0(s1,m3s,m4s) * 
  	 Cqaginfasqrtsing(U4a1kk,sqr(sqrt(m1s)+sqrt(m2s)),prec,&err1));
}

Cplx U4a1int(Double pse,Double m1se,Double m3se,Double m6se,
             Double m8se,Double prece,Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m3se; m3s=m6se; m4s=m8se; prec=prece;
  Double err;
  Cplx z=Cqaginf2(U4a1k,sqr(sqrt(m3s)+sqrt(m4s)),prec*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U4a2kk(Double s2)
{
  Cplx z;
  if(s1+s2 > ps*1e10)
    z = DM1DeltaB0(s2,m1s,m2s)*(ps*DB0z(s1,s2));
  else
    z = DM1DeltaB0(s2,m1s,m2s)*(B0A(ps,s1,s2) - B0A(0,s1,s2));
  return(z);
}

static Cplx U4a2k(Double x)
{
  Double err1=0;
  s1=x;
  return(DM1DeltaB0(s1,m3s,m4s) * 
  	 Cqaginfasqrtsing(U4a2kk,sqr(sqrt(m1s)+sqrt(m2s)),prec,&err1));
}

Cplx U4a2int(Double pse,Double m1se,Double m3se,Double m6se,
             Double m8se,Double prece,Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m6se; m3s=m3se; m4s=m8se; prec=prece;
// m1s and m3s are the masses of the doubled propagators!
  Double err;
  Cplx z=Cqaginf2(U4a2k,sqr(sqrt(m3s)+sqrt(m4s)),prec*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U4a3kk(Double s2)
{
  Cplx z;
  if(s1+s2 > ps*1e6)
    z = (m1s+m2s-s2)/(m1s*sla(s2,m1s,m2s))*(sqr(ps)/2*DDDM1B0z(s2,s1));
  else
    z = (m1s+m2s-s2)/(m1s*sla(s2,m1s,m2s))*(DM1B0(ps,s2,s1) - DM1B0(0,s2,s1)
    						- ps*DDM1B0z(s2,s1));
  return(z);
}

static Cplx U4a3k(Double x)
{
  Double err1=0;
  s1 = x;
  Cplx z = DeltaB0(s1,m3s,m4s) * 
  	 Cqaginfasqrtsing(U4a3kk,sqr(sqrt(m1s)+sqrt(m2s)),prec,&err1);
  return(z);
}

static Cplx U4a3k1(Double x)
{
  s1 = m5s-x;
  Double err1=0, thrsh=sqr(sqrt(m1s)+sqrt(m2s)), thrsh2=sqr(sqrt(ps)-sqrt(s1));
  Cplx z;
  if(thrsh2 > thrsh)
    z = DeltaB0(s1,m3s,m4s) * (Cqagasqrtsing(U4a3kk,thrsh,thrsh2,
    					      prec/min(thrsh2-thrsh,1.),&err1)
			       + Cqaginfasqrtsing(U4a3kk,thrsh2,prec,&err1));
  else
    z = DeltaB0(s1,m3s,m4s) * Cqaginfasqrtsing(U4a3kk,thrsh,prec,&err1);
  return(z);
}

Cplx U4a3int(Double pse,Double m1se,Double m3se,Double m6se,
             Double m8se,Double prece,Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m3se; m3s=m6se; m4s=m8se; prec=prece;
  Double err;
  Cplx z;
  m5s = sqr(sqrt(ps)-sqrt(m1s)-sqrt(m2s));
  if(m5s > sqr(sqrt(m3s)+sqrt(m4s)))
    z = Cqaginfasqrtsing2(U4a3k,m5s,prec*100,&err) +
         Cqag(U4a3k1,0,m5s-sqr(sqrt(m3s)+sqrt(m4s)),prec*100,&err);
  else
    z = Cqaginf2(U4a3k,sqr(sqrt(m3s)+sqrt(m4s)),prec*100,&err);
  *abserre = err/2;
  return(z/2);
}


static Cplx U5a1k1(Double s)
{
  Cplx z;
  if(s > ps*1e9)
    z = -DeltaB0(s,m3s,m4s)*real(B0A(s,m5s,m6s) - B0A(0,m5s,m6s)) * 
    	 (ps*DDM1B0(0,m2s,s));
  else
    z = -DeltaB0(s,m3s,m4s)*real(B0A(s,m5s,m6s) - B0A(0,m5s,m6s)) *
  	 (DM1B0(ps,m2s,s) - DM1B0(0,m2s,s));
  return(z);
}

static Cplx U5a1k2(Double s)
{
  Cplx z;
  if(s > ps*1e9)
    z = -DeltaB0(s,m5s,m6s)*real(B0A(s,m3s,m4s) - B0A(0,m3s,m4s)) * 
    	 (ps*DDM1B0(0,m2s,s));
  else
    z = -DeltaB0(s,m5s,m6s)*real(B0A(s,m3s,m4s) - B0A(0,m3s,m4s))*
  	 (DM1B0(ps,m2s,s) - DM1B0(0,m2s,s));
  return(z);
}

Cplx U5a1int(Double pse,Double m1se,Double m3se,Double m5se,
             Double m6se,Double m7se,Double prece,Double *abserre)
{
  ps=pse; m2s=m5se; m3s=m1se; m4s=m3se; m5s=m6se; m6s=m7se; prec=prece;
  Double err1, err2;
  Cplx z = Cqaginf(U5a1k1,sqr(sqrt(m3s)+sqrt(m4s)),prec,&err1)
          +Cqaginf(U5a1k2,sqr(sqrt(m5s)+sqrt(m6s)),prec,&err2);
  *abserre = sqrt(sqr(err1)+sqr(err2));
  return(z);
}


static Cplx U5a2k1(Double s)
{
  return(-DM1DeltaB0(s,m3s,m4s)*real(B0A(s,m5s,m6s) - B0A(0,m5s,m6s)) *
  	 (B0C(ps,m2s,s) - B0C(0,m2s,s)));
}

static Cplx U5a2k2(Double s)
{
  return(-DeltaB0(s,m5s,m6s)*real(DM1B0(s,m3s,m4s))*
  	 (B0C(ps,m2s,s) - B0C(0,m2s,s)));
}

Cplx U5a2int(Double pse,Double m1se,Double m3se,Double m5se,
             Double m6se,Double m7se,Double prece,Double *abserre)
{
  ps=pse; m2s=m5se; m3s=m1se; m4s=m3se; m5s=m6se; m6s=m7se; prec=prece;
  Double err1, err2;
  Cplx z = Cqaginf(U5a2k1,sqr(sqrt(m3s)+sqrt(m4s)),prec,&err1)
          +Cqaginf(U5a2k2,sqr(sqrt(m5s)+sqrt(m6s)),prec,&err2);
  *abserre = sqrt(sqr(err1)+sqr(err2));
  return(z);
}


#define DDeltaB0  DDeltaB0OnTwoPiI

static Cplx U5b1kk(Double s2)
{
  Cplx z;
  if(s1+s2 > ps*1e7)
    z = DeltaB0(s2,m3s,m4s)*ps*(DB0z(s1,s2) - DB0z(s1,m1s));
  else
    z = DeltaB0(s2,m3s,m4s)*(B0A(ps,s1,s2) - B0A(ps,s1,m1s) 
    			     - B0A(0,s1,s2) + B0A(0,s1,m1s));
  return(z);
}

static Cplx DU5b1kk(Double s2)
{
  Cplx z;
  if(s1+s2 > ps*1e7)
    z = DDeltaB0(s2,m3s,m4s)*ps*(DB0(0,s1,s2) - DB0(0,s1,m1s)) +
         DeltaB0(s2,m3s,m4s)*ps*DDM1B0(0,s2,s1);
  else
    z = DDeltaB0(s2,m3s,m4s)*(B0A(ps,s1,s2) - B0A(ps,s1,m1s) 
    			     - B0A(0,s1,s2) + B0A(0,s1,m1s)) +
         DeltaB0(s2,m3s,m4s)*(DM1B0(ps,s2,s1) - DM1B0(0,s2,s1));
  return(z);
}

static Cplx U5b1k(Double x)
{
  Double err1=0;
  s1=x;
  return(DeltaB0(s1,m5s,m6s) * 
         Cqagp2inf(U5b1kk,DU5b1kk,sqr(sqrt(m3s)+sqrt(m4s)),m1s,1,prec,&err1));
}

Cplx U5b1int(Double pse,Double m2se,Double m3se,Double m4se,
             Double m6se,Double m7se,Double prece,Double *abserre)
{
  ps=pse; m1s=m4se; m3s=m6se; m4s=m7se; m5s=m3se; m6s=m2se; prec=prece;
  Double err;
  Cplx z=Cqaginf2(U5b1k,sqr(sqrt(m5s)+sqrt(m6s)),prec*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U5b2kk(Double s2)
{
  Cplx z;
  if(s1+s2 > ps*1e7)
    z = DM1DeltaB0(s2,m5s,m6s)*(ps*DB0z(s1,s2));
  else
    z = DM1DeltaB0(s2,m6s,m5s)*(B0A(ps,s1,s2) - B0A(0,s1,s2));
  return(z);
}

static Cplx U5b2k(Double x)
{
  Double err1=0;
  s1=x;
  return(DeltaB0(s1,m3s,m4s) * 
  	 Cqaginf2(U5b2kk,sqr(sqrt(m5s)+sqrt(m6s)),prec,&err1));
}

Cplx U5b2int(Double pse,Double m2se,Double m3se,Double m4se,
             Double m6se,Double m7se,Double prece,Double *abserre)
{
  ps=pse; m1s=m4se; m3s=m6se; m4s=m7se; m5s=m3se; m6s=m2se; prec=prece;
  Double err;
  Cplx z=Cqagpinf(U5b2k,sqr(sqrt(m3s)+sqrt(m4s)),m1s,1,prec*100,&err);
  *abserre = err;
  return(z);
}


static Cplx U5c1kk(Double s2)
{
  Cplx z;
  z = DM1DeltaB0(s2,m5s,s1)*(B0C(ps,m2s,s2) - B0C(0,m2s,s2));
  return(z);
}

static Cplx U5c1k(Double x)
{
  Double err1=0;
  s1=x;
  return(DeltaB0(s1,m3s,m4s) * 
  	 Cqagpinf2(U5c1kk,sqr(sqrt(m5s)+sqrt(s1)),m1s,1,prec,&err1));
}

Cplx U5c1int(Double pse,Double m1se,Double m2se,Double m3se,Double m6se,
             Double m7se,Double prece,Double *abserre)
{
  ps=pse; m1s=m1se; m2s=m2se; m3s=m3se; m4s=m6se; m5s=m7se; prec=prece;
  Double err;
  Cplx z=Cqaginf(U5c1k,sqr(sqrt(m3s)+sqrt(m4s)),prec*100,&err);
  *abserre = err;
  return(z);
}
